#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

"""
This is the Station Client that reads measurements from sensors and reports them to the station monitoring service.
The station client is typically run on a raspberry-pi. One RPi represents one station.
"""
import logging
import sys
from logging.handlers import RotatingFileHandler

import requests
import urllib3

import auth
import constants
import settings
from sensors.bme680 import BME680
from sensors.sgp30 import SGP30
from sensors.sht31d import SHT31D
from sensors.sys import Sys
from sensors.tsl2591 import TSL2591

drivers = {
    constants.SENSOR_BME680: BME680,
    constants.SENSOR_SHT31_D: SHT31D,
    constants.SENSOR_TSL2591: TSL2591,
    constants.SENSOR_SGP30: SGP30,
    constants.SENSOR_SYS: Sys
}


def main():
    """
    Main entry point for the station client. Reads the connected sensors one by one and submits the measurements reports collected from
    these sensors to the station-monitoring-service.
    """
    try:
        _init_logger()
        logging.info('Running RCM Station Client (%s)', settings.station_name)
        # Disable warnings about unsecure connection, since HTTPS verify is set to False due to self signed certificate.
        urllib3.disable_warnings()
        for sensor in settings.sensors.split(','):
            if sensor not in drivers.keys():
                raise ValueError('Unsupported sensor: ' + sensor)
            driver = drivers[sensor]()
            measurement_report = driver.read()
            _submit(measurement_report)
    except Exception as exception:
        print('Unexpected error: ', exception)
        logging.exception('Unexpected error: ', exc_info=exception)
        sys.exit(1)


def _init_logger():
    logger = logging.getLogger()
    handler = RotatingFileHandler(settings.log_file, maxBytes=settings.log_max_bytes, backupCount=settings.log_backup_count)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(settings.log_level)


def _submit(measurement_report):
    logging.debug('Submitting measurement report for station %s to %s', settings.station_name, settings.reports_endpoint)
    payload = {
        'stationName': settings.station_name,
        'measurements': measurement_report
    }
    headers = {
        'content-type': 'application/json',
        'Authorization': f'Bearer {auth.get_token()}'
    }
    try:
        response = requests.post(settings.reports_endpoint, json=payload, headers=headers, verify=False, timeout=settings.timeout)
        response.raise_for_status()
    except Exception as exception:
        logging.exception('Failed to submit measurements report', exc_info=exception)
        sys.exit(1)


if __name__ == '__main__':
    main()



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

"""
Auth0 authentication
======================

When authentication is enabled, M2M access tokens are required in order for the station client to be able to submit measurement reports to
the station monitoring service.
When a token is requested, it will attempt to read the token from the cache and check if it hasn't expired yet.
Once the token is expired or if the token does not exist in cache, a request is made to Auth0 to obtain an new access token. The token is
written to a file along with the timestamp when it expires.
"""
import logging
import os.path
import sys
import time

import requests

import constants
import settings

_GRANT_TYPE_CLIENT_CREDENTIALS = 'client_credentials'
_ACCESS_TOKEN = 'access_token'
_EXPIRES_IN = 'expires_in'


def get_token():
    """
    Retrieve the access token for M2M authentication on the station monitoring service.
    Access tokens are 'cached' by writing them to a file locally.
    The file also contains the timestamp when the token expires. If the token was expired a new token will be retrieved.
    Returns None if authentication is disabled.
    """
    if not settings.auth_enabled:
        logging.debug('Skipping authentication because it is disabled')
        return None

    token = _read_token()
    if token is None:
        token = _request_token()

    return token


def _request_token():
    """
    Make request to the Auth0 server to request a new M2M access token.
    When a token is successfully obtained it will be stored in cache for future reference.
    :return: the freshly obtained access token.
    """
    logging.debug('Requesting new access token')
    payload = {
        'grant_type': _GRANT_TYPE_CLIENT_CREDENTIALS,
        'client_id': settings.client_id,
        'client_secret': settings.client_secret,
        'audience': settings.audience
    }

    try:
        response = requests.post(settings.token_endpoint, data=payload, timeout=settings.timeout)
        logging.debug('Received token-endpoint response: %s', response)
        response.raise_for_status()
        data = response.json()
        token = data[_ACCESS_TOKEN]
        expires_in = data[_EXPIRES_IN]
    except Exception as exception:
        logging.exception('Failed to request token', exc_info=exception)
        sys.exit(1)

    _write_token(token, expires_in)
    return token


def _read_token():
    """Read token from cache and return it. Returns None if token was not present or has expired."""
    logging.debug('Reading token from cache')
    if not os.path.isfile(settings.token_file):
        logging.debug('Token file does not exist')
        return None

    try:
        with open(settings.token_file, constants.MODE_READ) as file:
            content = file.read().splitlines()
            token = content[0]
            expires_at = float(content[1])
    except Exception as exception:
        logging.exception('Failed to read token from file', exc_info=exception)
        return None

    if expires_at <= time.time():
        logging.debug('Token expired')
        return None

    return token


def _write_token(access_token: str, expires_in: float):
    """
    Writes the token to the token file so it can be used for the next time until it expires.
    The access token is written to the first line. The timestamp when it expires is written to the second line.
    The timestamp is reduced with one minute so it will be renewed on time.

    :param access_token: The access token, excluding 'Bearer ' prefix
    :param expires_in: The number of seconds until the token will expire.
    """
    try:
        with open(settings.token_file, constants.MODE_WRITE) as file:
            token = access_token + constants.LINE_SEP
            expires_at = str(time.time() + expires_in - constants.SIXTY_SECONDS)
            file.writelines([token, expires_at])
    except Exception as exception:
        logging.exception('Failed to write token to file', exc_info=exception)
        sys.exit(1)



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

"""
System sensor that reports data of the system itself.
"""
import logging

import psutil

from sensors.sensor import Sensor


class Sys(Sensor):
    """Concrete implementation of the system sensor."""

    @staticmethod
    def _get_temps():
        """Returns a dict with temperature sensors and their current value"""
        if not hasattr(psutil, "sensors_temperatures"):
            print('Platform not supported')
            return {}
        temps = psutil.sensors_temperatures()
        if not temps:
            print('Could not read temperature sensors')
            return {}
        data = {}
        for name, entries in temps.items():
            for entry in entries:
                data[entry.label or name] = entry.current
        return data

    def _get_cpu_temp(self):
        temps = self._get_temps()
        if 'cpu_thermal' in temps:
            return temps['cpu_thermal']
        return None

    def read(self):
        """Read system metrics"""
        logging.info('Reading system sensor')

        disk = psutil.disk_usage('/')
        mem = psutil.virtual_memory()
        load = psutil.getloadavg()
        net = psutil.net_io_counters()
        cpu = psutil.cpu_count()
        cpu_freq = psutil.cpu_freq()

        measurements = {
            'disk-total': disk.total,
            'disk-free': disk.free,
            'disk-used': disk.used,
            'mem-total': mem.total,
            'mem-available': mem.available,
            'mem-used': mem.used,
            'load-avg-1': load[0],
            'load-avg-5': load[1],
            'load-avg-15': load[2],
            'net-bytes-sent': net.bytes_sent,
            'net-bytes-recv': net.bytes_recv,
            'net-packets-sent': net.packets_sent,
            'net-packets-recv': net.packets_recv,
            'net-errin': net.errin,
            'net-errout': net.errout,
            'net-dropin': net.dropin,
            'net-dropout': net.dropout,
            'cpu-count': cpu,
            'cpu-frequency': cpu_freq.current,
            'cpu-temp': self._get_cpu_temp(),
            'sys-boot-time': psutil.boot_time()
        }

        logging.info('Measurement: %s', measurements)
        return measurements



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

"""
The SGP30 sensor - tVOC and eCO2 sensor
=========================================

See https://www.adafruit.com/product/3709

The first time this sensor is used, the setting `sgp30-init-baseline` should be set to true and the script should be run in the background
for approx. 12 hours. This will establish a baseline for the sensor which is then written to disk. It will store the baseline every hour.
Check the SGP30 datasheet for more details.
See https://cdn-learn.adafruit.com/assets/assets/000/050/058/original/Sensirion_Gas_Sensors_SGP30_Datasheet_EN.pdf
"""
import logging
import os
import sys
import time

import adafruit_sgp30

from constants import MODE_WRITE, MODE_READ, LINE_SEP
from sensors.sensor import Sensor
from settings import init_baseline_sgp30, sgp30_baseline_file


class SGP30(Sensor):
    """
    Concrete implementation of the SGP30 sensor.
    """

    def __init__(self):
        """
        Initialize the sensor. After the sensor is initialized, it will be in an initialization phase for about 15 seconds
        and only return default values of eCO2 400 ppm and TVOC 0 ppb.
        """
        logging.info('Initializing SGP30 sensor')
        super().__init__()
        self._sensor = adafruit_sgp30.Adafruit_SGP30(self._i2c)

    def read(self):
        """
        Reads the sensor. The sensor is optimize for a sampling rate of 1Hz so we start reading measurements once every second for about
         30 seconds to get a stable reading and be sure initialization phase is completed.
        """
        logging.info('Reading SGP30 sensor')

        if init_baseline_sgp30:
            self._init_baseline()

        self._apply_baseline()

        elapsed_sec = 0
        while elapsed_sec < 30:
            self._compile_measurement_report()
            time.sleep(1)
            elapsed_sec += 1

        report = self._compile_measurement_report()
        logging.info('Measurement: %s', report)
        return report

    def _compile_measurement_report(self):
        """
        Send a measurement command to the sensor and read the values to compile the measurement report and return it.
        """
        self._sensor.iaq_measure()
        measurements = {
            'raw-ethanol': self._sensor.Ethanol,
            'raw-H2': self._sensor.H2,
            'TVOC': self._sensor.TVOC,
            'eCO2': self._sensor.eCO2,
        }
        logging.debug('Measurement: %s', measurements)
        return measurements

    def _apply_baseline(self):
        """Read the baseline values from disk and configure the sensor with these values"""
        logging.debug('Reading IAQ baseline values from file')
        if not os.path.isfile(sgp30_baseline_file):
            logging.warning('SGP30 baseline values not stored yet, please establish baseline first!')
            return

        try:
            with open(sgp30_baseline_file, MODE_READ) as file:
                content = file.read().splitlines()
                eco2_baseline = int(content[0])
                tvoc_baseline = int(content[1])
                self._sensor.set_iaq_baseline(eco2_baseline, tvoc_baseline)
                logging.info('Applied baseline: eCO2 = 0x%x, TVOC = 0x%x', eco2_baseline, tvoc_baseline)
        except Exception as exception:
            logging.exception('Failed to read SGP30 baseline values from file', exc_info=exception)
            sys.exit(1)

    def _init_baseline(self):
        """
        Start reading the sensors continuously to establish the baseline. Best baseline is achieved after approx. 12 hours.
        """
        logging.warning('Initializing IAQ baseline, please let it run for at least 12 hours')
        elapsed_sec = 0
        while True:
            self._compile_measurement_report()
            time.sleep(1)
            elapsed_sec += 1
            if elapsed_sec > 60:
                logging.info('Baseline: eCO2 = 0x%x, TVOC = 0x%x', self._sensor.baseline_eCO2, self._sensor.baseline_TVOC)
            if elapsed_sec > 3600:
                elapsed_sec = 0
                self._store_baseline()

    def _store_baseline(self):
        """Write the baseline values to disk"""
        logging.debug('Writing IAQ baseline values to file')
        try:
            with open(sgp30_baseline_file, MODE_WRITE) as file:
                eco2_baseline = str(self._sensor.baseline_eCO2) + LINE_SEP
                tvoc_baseline = str(self._sensor.baseline_TVOC) + LINE_SEP
                timestamp = str(time.time())
                file.writelines([eco2_baseline, tvoc_baseline, timestamp])
        except Exception as exception:
            logging.exception('Failed to write SGP30 baseline values to file', exc_info=exception)
            sys.exit(1)



#  Copyright (c) 2020 Pim Hazebroek
#  This program is made available under the terms of the MIT License.
#  For full details check the LICENSE file at the root of the project.

"""
The SHT31-D sensor - High precision Temperature & Humidity sensor
=================================================================

See https://www.adafruit.com/product/2857
"""
import logging

import adafruit_sht31d

from sensors.sensor import Sensor


class SHT31D(Sensor):
    """
    Concrete implementation of SHT31-D sensor.
    """
    def read(self):
        """
        Reads the temperature and relative humidity.
        """
        logging.info('Reading SHT31-D sensor')
        sensor = adafruit_sht31d.SHT31D(self._i2c)
        measurements = {
            'temperature': sensor.temperature,
            'humidity': sensor.relative_humidity
        }
        logging.info('Measurement: %s', measurements)
        return measurements
